@@ -6,7 +6,11 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
+
+#if !defined(_MSC_VER)
 #include <unistd.h>
+#endif
+
 #include <stdlib.h>
 #include <string.h>
 #include "sqlite3.h"
